@@ -17,6 +17,7 @@
 
 #include <apt-private/private-cmndline.h>
 
+#include <array>
 #include <cstdio>
 #include <iostream>
 #include <memory>
@@ -25,7 +26,7 @@
 #include <sys/types.h>
 #include <sys/wait.h>
 
-#include <string.h>
+#include <cstring>
 #include <unistd.h>
 
 #include <apti18n.h>
@@ -99,6 +100,7 @@ int main(int argc,const char *argv[])					/*{{{*/
    FileFd forward;
    if (is_forwarding_dumper)
    {
+      signal(SIGPIPE, SIG_IGN);
       int external[] = {-1, -1};
       if (pipe(external) != 0)
 	 return 250;
@@ -133,11 +135,10 @@ int main(int argc,const char *argv[])					/*{{{*/
       return WriteError("ERR_READ_ERROR", out, stdoutfd, Solver);
    }
 
-   constexpr size_t BufSize = 64 * 1024;
-   std::unique_ptr<char[]> Buf(new char[BufSize]);
+   std::array<char, APT_BUFFER_SIZE> Buf;
    unsigned long long ToRead = 0;
    do {
-      if (input.Read(Buf.get(),BufSize, &ToRead) == false)
+      if (input.Read(Buf.data(), Buf.size(), &ToRead) == false)
       {
 	 std::ostringstream out;
 	 out << "Writing EDSP solver input to file '" << filename << "' failed as reading from stdin failed!\n";
@@ -145,22 +146,15 @@ int main(int argc,const char *argv[])					/*{{{*/
       }
       if (ToRead == 0)
 	 break;
-      if (forward.IsOpen() && forward.Failed() == false && forward.Write(Buf.get(),ToRead) == false)
+      if (forward.IsOpen() && forward.Failed() == false && forward.Write(Buf.data(),ToRead) == false)
 	 forward.Close();
-      if (dump.IsOpen() && dump.Failed() == false && dump.Write(Buf.get(),ToRead) == false)
+      if (dump.IsOpen() && dump.Failed() == false && dump.Write(Buf.data(),ToRead) == false)
 	 dump.Close();
    } while (true);
    input.Close();
    forward.Close();
    dump.Close();
 
-   if (_error->PendingError())
-   {
-      std::ostringstream out;
-      out << "Writing EDSP solver input to file '" << filename << "' failed due to write errors!\n";
-      return WriteError("ERR_WRITE_ERROR", out, stdoutfd, Solver);
-   }
-
    if (is_forwarding_dumper)
    {
       // Wait and collect the error code
@@ -179,6 +173,12 @@ int main(int argc,const char *argv[])					/*{{{*/
       else
 	 return 255;
    }
+   else if (_error->PendingError())
+   {
+      std::ostringstream out;
+      out << "Writing EDSP solver input to file '" << filename << "' failed due to write errors!\n";
+      return WriteError("ERR_WRITE_ERROR", out, stdoutfd, Solver);
+   }
    else
       EDSP::WriteError("ERR_JUST_DUMPING", "I am too dumb, i can just dump!\nPlease use one of my friends instead!", stdoutfd);
    return 0;
